"""
Kademlia is a Python implementation of the Kademlia protocol for U{Twisted <http://twistedmatrix.com/trac/>}.

@author: Brian Muller U{bamuller@gmail.com}
"""
version_info = (0, 2)
version = '.'.join(map(str, version_info))
